// DirectX.cpp: implementation of the CDirectX class.
//
//////////////////////////////////////////////////////////////////////

#include "Snes9X.h"
#include "DirectX.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CDirectX::CDirectX()
{
	lpDD = NULL;
	lpDDClipper = NULL;
	lpDDPalette = NULL;

	lpDDSPrimary = NULL;
	lpDDSBack = NULL;
	lpDDSOffScreen = NULL;

	lpDDSPrimary2 = NULL;
	lpDDSBack2 = NULL;
	lpDDSOffScreen2 = NULL;

	Width = Height = -1;
	Depth = -1;
	Windowed = true;
	DoubleBuffered = false;
	fpsShow = false;

	lpDS = NULL;
	lpDSB = NULL;
	lpDSBPrimary = NULL;
	
	DSAvailable = false;
	for( short C = 0; C != 15; C++)
		Till[C] = timeGetTime();
}

CDirectX::~CDirectX()
{
	DeInitialize();
}

bool CDirectX::Initialize()
{
	dErr = DirectDrawCreate( NULL, &lpDD, NULL);
	if( dErr != DD_OK)
		return false;

	dErr = lpDD -> CreateClipper( 0, &lpDDClipper, NULL);
    if( dErr != DD_OK)
		return false;

	dErr = lpDDClipper -> SetHWnd( 0, GUI.hWnd);
    if( dErr != DD_OK)
		return false;

	dErr = DirectSoundCreate( NULL, &lpDS, NULL);
	if( dErr != DS_OK)
	{	DSAvailable = false;	}
	else
	{
		dErr = lpDS -> SetCooperativeLevel( GUI.hWnd, DSSCL_EXCLUSIVE);
		if( dErr != DS_OK)
		{
			lpDS -> Release();
			lpDS = NULL;

			DSAvailable = false;
		}
		else
		{	DSAvailable = true;	}
	}

	hFont = CreateFont( 12, 0, 0, 0,
				FW_NORMAL, FALSE, FALSE, FALSE,
				ANSI_CHARSET, 
				OUT_DEFAULT_PRECIS,
				CLIP_DEFAULT_PRECIS,
				NONANTIALIASED_QUALITY,
				VARIABLE_PITCH,
				"MS Sans Serif");

	return true;
}

void CDirectX::DeInitialize()
{
	if( lpDD != NULL)
	{
		if( lpDDSPrimary != NULL)
		{
			lpDDSPrimary -> Release();
			lpDDSPrimary = NULL;
		}
		if( lpDDSOffScreen != NULL)
		{
			lpDDSOffScreen -> Release();
			lpDDSOffScreen = NULL;
		}
		if( lpDDSPrimary2 != NULL)
		{
			lpDDSPrimary2 -> Release();
			lpDDSPrimary2 = NULL;
		}
		if( lpDDSOffScreen2 != NULL)
		{
			lpDDSOffScreen2 -> PageUnlock(0);
			lpDDSOffScreen2 -> Release();
			lpDDSOffScreen2 = NULL;
		}
		if( lpDDClipper != NULL)
		{
			lpDDClipper -> Release();
			lpDDClipper = NULL;
		}
		if( lpDDPalette != NULL)
		{
			lpDDPalette -> Release();
			lpDDPalette = NULL;
		}
		lpDD -> Release();
		lpDD = NULL;
	}

	if( lpDS != NULL)
	{
		if( lpDSB != NULL)
		{
			lpDSB -> Release();
			lpDSB = NULL;
		}
		if( lpDSBPrimary != NULL)
		{
			lpDSBPrimary -> Release();
			lpDSBPrimary = NULL;
		}
	}
}

bool CDirectX::SetDisplayMode( 
		int pWidth, int pHeight, 
		char pDepth, bool pWindowed, bool pDoubleBuffered)
{
DDSURFACEDESC ddsd;
DDSCAPS ddscaps;
PALETTEENTRY PaletteEntries[ 256];

	if( pWidth == Width && pHeight == Height &&
		(pDepth == Depth || pDepth == 0) && pWindowed == Windowed && 
		pDoubleBuffered == pDoubleBuffered)
		return true;

	if( pDepth == 0)
		pDepth = Depth;

	if( lpDDSPrimary != NULL)
	{
		lpDDSPrimary -> Release();
		lpDDSPrimary = NULL;
	}
	if( lpDDSOffScreen != NULL)
	{
		lpDDSOffScreen -> Release();
		lpDDSOffScreen = NULL;
	}
	if( lpDDSPrimary2 != NULL)
	{
		lpDDSPrimary2 -> Release();
		lpDDSPrimary2 = NULL;
	}
	if( lpDDSOffScreen2 != NULL)
	{
		lpDDSOffScreen2 -> PageUnlock(0);
		lpDDSOffScreen2 -> Release();
		lpDDSOffScreen2 = NULL;
	}
	if( lpDDPalette != NULL)
	{
		lpDDPalette -> Release();
		lpDDPalette = NULL;
	}

	lpDD -> FlipToGDISurface();

	if( pWindowed)
	{	
		lpDD -> RestoreDisplayMode();

		SetWindowLong( GUI.hWnd, GWL_STYLE, WS_POPUPWINDOW|WS_CAPTION|WS_THICKFRAME|WS_VISIBLE);
		SetWindowPos( GUI.hWnd, HWND_TOP, 0, 0, 0, 0, SWP_DRAWFRAME|SWP_FRAMECHANGED|SWP_NOMOVE|SWP_NOSIZE);

		ZeroMemory( &ddsd, sizeof( ddsd));

		ddsd.dwSize = sizeof( ddsd);
		ddsd.dwFlags = DDSD_PIXELFORMAT;
		dErr = lpDD -> GetDisplayMode( &ddsd);
		if( dErr != DD_OK)
		{	pDepth = 8;	}
		else
		{
			if( ddsd.ddpfPixelFormat.dwFlags&DDPF_RGB)
			{	pDepth = (char) ddsd.ddpfPixelFormat.dwRGBBitCount;	}
			else
			{	pDepth = 8;	}
		}

		if( pDepth == 8)
		{	dErr = lpDD -> SetCooperativeLevel( GUI.hWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN|DDSCL_ALLOWREBOOT);
		}
		else
		{	dErr = lpDD -> SetCooperativeLevel( GUI.hWnd, DDSCL_NORMAL|DDSCL_ALLOWREBOOT);	}
	} 
	else 
	{
		SetWindowLong( GUI.hWnd, GWL_STYLE, WS_POPUP|WS_VISIBLE);
		SetWindowPos( GUI.hWnd, HWND_TOP, 0, 0, 0, 0, SWP_DRAWFRAME|SWP_FRAMECHANGED|SWP_NOMOVE|SWP_NOSIZE);

		if( Windowed || pWidth != Width || pHeight != Height || pDepth != Depth)
		{	
			dErr = lpDD -> SetCooperativeLevel( GUI.hWnd, DDSCL_ALLOWMODEX|DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN|DDSCL_ALLOWREBOOT);

			if( dErr == DD_OK)
				dErr = lpDD -> SetDisplayMode( pWidth, pHeight, pDepth);	
		}
	}
	if( dErr != DD_OK)
		return false;

	ZeroMemory( &ddsd, sizeof( ddsd));
	if( !pWindowed && pDoubleBuffered)
	{
		ddsd.dwSize = sizeof( ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
		ddsd.dwBackBufferCount = 1;
	}
	else
	{
		ddsd.dwSize = sizeof( ddsd);
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	}

	dErr = lpDD -> CreateSurface( &ddsd, &lpDDSPrimary, NULL);
    if( dErr != DD_OK )
	{
		if( pDoubleBuffered)
		{
			ddsd.dwSize = sizeof( ddsd);
			ddsd.dwFlags = DDSD_CAPS;
			ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

			pDoubleBuffered = false;

			dErr = lpDD -> CreateSurface( &ddsd, &lpDDSPrimary, NULL);
    	}

		if( dErr != DD_OK)
			return false;
	}

	ZeroMemory( &DDPixelFormat, sizeof( DDPixelFormat));
	DDPixelFormat.dwSize = sizeof( DDPixelFormat);
	lpDDSPrimary -> GetPixelFormat( &DDPixelFormat);

	if( pWindowed)
	{	lpDDSPrimary -> SetClipper( lpDDClipper);	}

    if( FAILED( lpDDSPrimary -> QueryInterface( IID_IDirectDrawSurface2, (void **)&lpDDSPrimary2)))
    {   lpDDSPrimary2 = NULL;  }
    else
    {
		lpDDSPrimary -> Release();
		lpDDSPrimary = NULL;
    }

	if( pDoubleBuffered)
	{
	RECT Src;
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;

		if( lpDDSPrimary2 != NULL)
		{	dErr = lpDDSPrimary2 -> GetAttachedSurface(&ddscaps, &lpDDSBack2);	}
		else
		{	dErr = lpDDSPrimary -> GetAttachedSurface(&ddscaps, &lpDDSBack);	}
		if( dErr != DD_OK )
			return false;

		Src.left = 0; Src.right = pWidth;
		Src.top = 0; Src.bottom = pHeight;
		if( lpDDSPrimary2 != NULL)
		{	lpDDSBack2 -> BltFast( 0, 0, lpDDSPrimary2, &Src, DDBLTFAST_WAIT);	}
		else
		{	lpDDSBack -> BltFast( 0, 0, lpDDSPrimary, &Src, DDBLTFAST_WAIT);	}
	}
	else
	{	
		lpDDSBack = lpDDSPrimary;	
		lpDDSBack2 = lpDDSPrimary2;
	}

	ddsd.dwSize = sizeof( ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
	ddsd.dwWidth = 256 * 2;
	ddsd.dwHeight = 239 * 2;
	
	dErr = lpDD -> CreateSurface( &ddsd, &lpDDSOffScreen, NULL);
    if( dErr != DD_OK )
		return false;

	if( lpDDSPrimary2 != NULL)
    if( FAILED( lpDDSOffScreen -> QueryInterface( IID_IDirectDrawSurface2, (void **)&lpDDSOffScreen2)))
    {   lpDDSOffScreen2 = NULL;  }
    else
    {
		lpDDSOffScreen2 -> PageLock(0);
		lpDDSOffScreen -> Release();
		lpDDSOffScreen = NULL;
    }

	if( pDepth == 8)
	{
		dErr = lpDD -> CreatePalette( DDPCAPS_8BIT|DDPCAPS_ALLOW256, PaletteEntries, &lpDDPalette, NULL);
		if( dErr != DD_OK)
		{
			lpDDPalette = NULL;
			return false;
		}
	}

	Windowed = pWindowed;
	Depth = pDepth;
	Height = pHeight;
	Width = pWidth;
	DoubleBuffered = pDoubleBuffered;
	return true;
}

extern long _rate;
extern bool _sixteen_bit;
extern bool _stereo;
extern long _buffernos;
extern long _buffersize;
extern long _samplecount;
extern long _bytecount;
extern long _interval;

bool CDirectX::SetSoundMode( void)
{
    extern FILE *trace_fs;

    DSBUFFERDESC dsbd;

    if( lpDSB != NULL)
    {
        lpDSB -> Release();
        lpDSB = NULL;
    }
    if( lpDSBPrimary != NULL)
    {
        lpDSBPrimary -> Release();
        lpDSBPrimary = NULL;
    }

    if( _rate == 0)
    {
        fprintf (trace_fs, "_rate is 0\n");
        fflush (trace_fs);
        return false;
    }

    if( !DSAvailable)
    {
        fprintf (trace_fs, "!DSAvailable in DirectX::SetSoundMode\n");
        fflush (trace_fs);
        return false;
    }

    wfx.wFormatTag = WAVE_FORMAT_PCM;
    wfx.nChannels = _stereo ? 2 : 1;
    wfx.nSamplesPerSec = _rate;
    wfx.nBlockAlign = (_sixteen_bit ? 2 : 1) * (_stereo ? 2 : 1);
    wfx.wBitsPerSample = _sixteen_bit ? 16 : 8;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
    wfx.cbSize = 0;

    ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
    dsbd.dwSize = sizeof(dsbd);
    dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;

    dErr = lpDS -> CreateSoundBuffer( &dsbd, &lpDSBPrimary, NULL);
    if( dErr != DS_OK)
    {
        fprintf (trace_fs, "Create primary buffer failed\n");
        fflush (trace_fs);
        lpDSB = NULL;
        return false;
    }

    lpDSBPrimary -> SetFormat( &wfx);
    lpDSBPrimary -> Play (0, 0, DSBPLAY_LOOPING);
    ZeroMemory( &dsbd, sizeof( dsbd));
    dsbd.dwSize = sizeof( dsbd);
    dsbd.dwFlags = DSBCAPS_CTRLDEFAULT | DSBCAPS_GETCURRENTPOSITION2;
    dsbd.dwBufferBytes = _buffersize * _buffernos;
    dsbd.lpwfxFormat = &wfx;

    dErr = lpDS -> CreateSoundBuffer( &dsbd, &lpDSB, NULL);
    if( dErr != DS_OK)
    {
        fprintf (trace_fs, "CreateSoundBuffer for secondary failed\n");
        fflush (trace_fs);
        lpDSBPrimary -> Release ();
        return false;
    }
    return true;
}

bool CDirectX::LockSurface( BYTE **lpSurface, DWORD *lpPitch)
{
DDSURFACEDESC ddsd;
	
	*lpSurface = NULL;

	ddsd.dwSize = sizeof( ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	while( 1)
	{
		if( lpDDSOffScreen2 != NULL)
		{	dErr = lpDDSOffScreen2 -> Lock( NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL);	}
		else
		{	dErr = lpDDSOffScreen -> Lock( NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL);	}

		if( dErr == DD_OK)
			break;

		if( dErr == DDERR_SURFACELOST)
		{
			if( lpDDSOffScreen2 != NULL)
			{	dErr = lpDDSOffScreen2 -> Restore();	}
			else
			{	dErr = lpDDSOffScreen -> Restore();	}
			if( dErr != DD_OK)
				break;

			continue;
		}

		if( dErr != DDERR_WASSTILLDRAWING)
			break;
	}
	if( dErr != DD_OK)
		return false;

	*lpSurface = (BYTE *) ddsd.lpSurface;
	*lpPitch = ddsd.lPitch;
	return true;
}

bool CDirectX::UnlockSurface( void)
{
	if( lpDDSOffScreen2 != NULL)
	{	lpDDSOffScreen2 -> Unlock( NULL);	}
	else
	{	lpDDSOffScreen -> Unlock( NULL);	}
	return true;
}

#include "snes9x/snes9x.h"
bool CDirectX::BltSurface( int pWidth, int pHeight)
{
RECT dst, src;
POINT p;
char S[100];
int C;
long CT = timeGetTime();

	S[0] = '\0';

	if( CT < Till[0])
	{	strcpy( S, Msg[0]);	}
	else
	{
		for( short I = 0, L = 0; I != 15; I ++)
		if( CT < Till[I])
		{
			Till[L] = Till[I];
			strcpy( Msg[L], Msg[I]);
			L ++;
		}
		for( ; L != 15; L ++)
			Till[L] = CT;

		if( CT < Till[0])
			strcpy( S, Msg[0]);
	}
	
	if( fpsShow)
	{
	double fps = 1000;

		fps /= CT-fpsLast;

		for( C = 0; C != 99; C++) fpsAvg[C] = fpsAvg[C+1];
		fpsAvg[99] = fps;

		fps = 0;
		for( C = 0; C != 100; C++) fps += fpsAvg[C];

		fps /= 100;

		if( S[0] != '\0')
		{	sprintf( S, "%s [%0.2f fps, %i ms]", S, fps, CT-fpsLast);	}
		else
		{	sprintf( S, "%0.2f fps, %i ms", fps, CT-fpsLast);	}

		fpsLast = CT;
	}

	if( S[0] != '\0')
	{
	HDC hDC;
		if( lpDDSOffScreen2 != NULL)
		{	dErr = lpDDSOffScreen2 -> GetDC( &hDC);	}
		else
		{	dErr = lpDDSOffScreen -> GetDC( &hDC);	}

		if( dErr == DD_OK)
		{
		RECT rc;

			SelectObject(hDC, hFont);
			SetTextColor(hDC, RGB(255,255,255));
			SetBkMode(hDC, TRANSPARENT);

			SetRect(&rc, 0, pHeight-(12*15), pWidth, pHeight);
			for( short I = 0, D = 0; I != 15; I ++)
			{
			char O[100];
				if( I == 0)
				{	strcpy( O, S);	}
				else
				{	
					O[0] = '\0';
					if( CT < Till[I])
						strcpy( O, Msg[I]);
				}

				if( O[0] != '\0')
				{
					ExtTextOut(hDC, 0, pHeight-(12*(D+1)), 0, &rc, O, strlen( O), NULL);
					D++;
				}
			}

			if( lpDDSOffScreen2 != NULL)
			{	dErr = lpDDSOffScreen2 -> ReleaseDC( hDC);	}
			else
			{	dErr = lpDDSOffScreen -> ReleaseDC( hDC);	}
		}
	}

	src.left = 0;
	src.right = pWidth;
	src.top = 0;
	src.bottom = pHeight;

	if( Settings.SupportHiRes)
	{
		if( pWidth <= 256)
			pWidth <<= 1;

		if( pHeight <= 239)
			pHeight <<= 1;
	}

	if( Windowed)
	{
		p.x = p.y = 0;
		ClientToScreen( GUI.hWnd, &p);

		dst.left = p.x;
		dst.right = p.x + Width;
		dst.top = p.y;
		dst.bottom = p.y + Height;
	}
	else
	{
		dst.left = (Width-pWidth)>>1;
		if( dst.left <= 0)
			dst.left = 0;
		dst.right = dst.left + pWidth;

		dst.top = (Height-pHeight)>>1;
		if( dst.top <= 0)
			dst.top = 0;
		dst.bottom = dst.top + pHeight;
	}

	if( lpDDSOffScreen2 != NULL)
	if( lpDDSBack2 -> Blt( &dst, lpDDSOffScreen2, &src, DDBLT_WAIT, NULL) == DD_OK)
	{	return true;	}
	else
	{	return false;	}


	if( lpDDSOffScreen != NULL)
	if( lpDDSBack -> Blt( &dst, lpDDSOffScreen, &src, DDBLT_WAIT, NULL) == DD_OK)
	{	return true;	}
	else
	{	return false;	}

	return false;
}

void CDirectX::OnScreenMsg(char * pMsg, int pDur)
{
	for( short C = 0; C != 15; C ++)
	if( timeGetTime() >= (unsigned long) Till[C])
	{
		strcpy( Msg[C], pMsg);
		Till[C] = timeGetTime() + pDur;
		break;
	}
}
